use std::{
    cell::{Cell, RefCell},
    collections::HashMap,
    os::unix::prelude::RawFd,
    rc::Rc,
};

use crate::prelude::*;

mod udp;
pub use udp::UdpSocketBuilder;
mod timerfd;
pub use self::timerfd::TimerFdBuilder;

pub struct InterestRegistration {
    poll: Rc<RefCell<mio::Poll>>,
    token: mio::Token,
    fd: RawFd,
    interest: Cell<mio::Interest>,
}

impl InterestRegistration {
    pub fn update_interest(&self, to: mio::Interest) {
        if self.interest.get() != to {
            self.poll
                .borrow()
                .registry()
                .reregister(&mut mio::unix::SourceFd(&self.fd), self.token, to)
                .expect("couldn't update FD interest");
            self.interest.set(to);
        }
    }
}

pub trait FdHandler {
    fn ready_read(&self, rc: &Rc<dyn FdHandler>);
    fn ready_write(&self, rc: &Rc<dyn FdHandler>);
    fn dispatch(&self, msg: super::msg::Message);
}

pub struct IOService {
    api: Rc<crate::API>,
    poll: Rc<RefCell<mio::Poll>>,
    next_token: Cell<mio::Token>,
    handlers: RefCell<HashMap<mio::Token, Rc<dyn FdHandler>>>,
}

impl Service for IOService {
    fn new(api: Rc<crate::API>) -> Self {
        Self {
            api,
            poll: Rc::new(RefCell::new(
                mio::Poll::new().expect("couldn't create poll?"),
            )),
            next_token: Cell::new(mio::Token(1)),
            handlers: Default::default(),
        }
    }
    fn setup(self: &Rc<Self>) {
        self.api.create_channel::<super::ReceivePacketChannel>();
        self.api.create_channel::<super::SendPacketChannel>();
    }
}

impl IOService {
    pub fn register_handler(&self, reg: &InterestRegistration, handler: Rc<dyn FdHandler>) {
        self.handlers.borrow_mut().insert(reg.token, handler);
    }

    pub fn register_interest(&self, fd: RawFd) -> InterestRegistration {
        let token = self.next_token.get();
        self.next_token.set(mio::Token(token.0 + 1));

        self.poll
            .borrow()
            .registry()
            .register(
                &mut mio::unix::SourceFd(&fd),
                token,
                mio::Interest::READABLE,
            )
            .expect("couldn't register initial readable interest?");

        InterestRegistration {
            poll: self.poll.clone(),
            token,
            fd,
            interest: Cell::new(mio::Interest::READABLE),
        }
    }

    pub(crate) fn run(&self) {
        let mut events = mio::Events::with_capacity(128);
        loop {
            let pr = self.poll.borrow_mut().poll(&mut events, None);
            if pr.is_err() {
                continue;
            }
            let handlers = self.handlers.borrow();
            for evt in &events {
                let h = handlers.get(&evt.token()).unwrap();
                if evt.is_readable() {
                    h.ready_read(h);
                }
                if evt.is_writable() {
                    h.ready_write(h);
                }
            }

            self.api.services.borrow().event_root().fire_all();
        }
    }
}
